/* 

   Demo Program for simple freeglut init


   (c) 2012  Ivo Ihrke / Tobias RItschel
       Universitaet des Saarlandes
       ihrke@mmci.uni-saarland.de
 
*/

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>

#ifdef WIN32
#define _USE_MATH_DEFINES
#endif
#include <math.h>
#include <time.h>

#ifndef WIN32
#include <unistd.h>
#endif

// Include GL extension header
#include <GL/glew.h>

// Include freeglut windoing
#include <GL/freeglut.h>

// Our image class
#include "../Image/pvc_image.h"

using namespace std;

// Load an image from a file with a name & erro check
static PVC_Image* loadImage(const string& filename) {
  PVC_Image* const result = new PVC_Image();

  if (result-> load( filename.c_str() ) == 0) {
    cout << "The file '" << filename << "' has a size of " << result->w() << "x" << result->h() << endl;
  } else {
    cerr << "The file '" <<filename << "' could not be loaded ..." << endl;
  }
  return result;
}

// Load a file, e.g. a shader as a string
string loadFileAsString(const string& fileName) {
    
	ifstream ifs(fileName.c_str(), ios::in | ios::binary | ios::ate);

    ifstream::pos_type fileSize = ifs.tellg();
    ifs.seekg(0, ios::beg);

    vector<char> bytes((unsigned int)fileSize);
    ifs.read(&bytes[0], fileSize);

    return string(&bytes[0], (unsigned int)fileSize);
}

// Slight GL abstraction

// --------

// Load an image into a texture and set the minimal state
static GLuint loadTexture(const PVC_Image* const pvc_image) {

	GLuint texture;
	glGenTextures(1, &texture);

	glBindTexture(GL_TEXTURE_2D, texture);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, pvc_image->w(), pvc_image->h(), 0, GL_BGRA, GL_FLOAT, pvc_image->get_data());

	return texture;
}

// Allocates space for a texture without actually putting new data into the texture image
static GLuint allocateTexture(const int width, const int height, const GLint internalFormat = GL_RGBA8, const bool useMIPmap = false) {
	GLuint texture;
	glGenTextures(1, &texture);

	glBindTexture(GL_TEXTURE_2D, texture);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, width, height, 0, GL_BGRA, GL_FLOAT, NULL);

	if(useMIPmap) {
		glGenerateMipmap(GL_TEXTURE_2D);
	}

	return texture;
}

static bool fileExists(const string& filename) {
	ifstream file(filename.c_str(), std::ios::in);
	return !file.fail();
}

static void dumpInfoLog(const string& infoLog) {
	cerr << "------------------------------------" << endl;
	cerr << "------------------------------------" << endl;
	cerr << infoLog;
	cerr << "------------------------------------" << endl;
	cerr << "------------------------------------" << endl;
}

static string getShaderInfoLog(const GLuint shader) {
	const int maxInfoLogSize = 1000;	
	GLchar infoLog[maxInfoLogSize];
	GLint infoLogSize;
	glGetShaderInfoLog(shader, maxInfoLogSize, &infoLogSize, infoLog);
	return infoLog;
}

static string getProgramInfoLog(const GLuint program) {
	const int maxInfoLogSize = 1000;	
	GLchar infoLog[maxInfoLogSize];
	GLint infoLogSize;
	glGetProgramInfoLog(program, maxInfoLogSize, &infoLogSize, infoLog);
	return infoLog;
}

static bool didCompile(const GLhandleARB shader) {
	int param;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &param);
	return GL_TRUE  == param;
}

static bool didLink(const GLhandleARB program) {
	int param;
	glGetProgramiv(program, GL_LINK_STATUS, &param);
	return GL_TRUE  == param;
}

// Load a single shader (faragment or vertex or ..) from a file
static GLhandleARB loadShader(const string& filename, const GLenum shaderType) {
	const GLhandleARB shader = glCreateShaderObjectARB(shaderType);

	const string source = loadFileAsString(filename);
	const GLcharARB* source2 = source.c_str();

	glShaderSourceARB(shader, 1, &source2, NULL);

	glCompileShader(shader);

	if(!didCompile(shader)) {
		//cout << "The shader '" << filename << "' did not compile successfully" << endl;		
		//dumpInfoLog(getShaderInfoLog(shader));
	}

	return shader;
}

static void attachShader(const string& filename, const GLuint program, const GLenum shaderType) {
	if(fileExists(filename)) {
		glAttachObjectARB(program, loadShader(filename, shaderType));
	}
}

// Load a "shader" (to be precise: a program with a fragment shader) using a naming convention & error checking
static GLuint loadProgram(const string& filename) {
    
	const GLuint program = glCreateProgramObjectARB();

	attachShader(filename + "Fragment.glsl", program, GL_FRAGMENT_SHADER);
	attachShader(filename + "Vertex.glsl", program, GL_VERTEX_SHADER);
	attachShader(filename + "Geometry.glsl", program, GL_GEOMETRY_SHADER);

	glLinkProgramARB(program);	

	if(!didLink(program)) {
		cout << "The program '" << filename << "' did not link successfully" << endl;		
		dumpInfoLog(getProgramInfoLog(program));
	}

	return program;
}

// Wrapper funtions to set uniform values more easily

static void setUniformFloat(const GLuint program, const string& name, const float x) {
	const GLint location = glGetUniformLocation(program, name.c_str());
	const float values[] = {x};
	glUniform1fv(location, 1, values);
}

static void setUniformInt(const GLuint program, const string& name, const int x) {
	const GLint location = glGetUniformLocation(program, name.c_str());
	const int values[] = {x};
	glUniform1iv(location, 1, values);
}

static void setUniformFloat2(const GLuint program, const string& name, const float x, const float y) {
	const GLint location = glGetUniformLocation(program, name.c_str());
	const float values[] = {x,y};
	glUniform2fv(location, 1, values);
}

static void setUniformFloat3(const GLuint program, const string& name, const float x, const float y, const float z) {
	const GLint location = glGetUniformLocation(program, name.c_str());
	const float values[] = {x,y,z};
	glUniform3fv(location, 1, values);
}

// ------

static int windowWidth;
static int windowHeight;

// Some GL objects

static GLuint texture;
static GLuint lowresTexture;
static GLuint fboTestTexture;
static GLuint histogramTexture;
static GLuint histoPyramidTexture;
static GLuint resultHistoPyramidTexture;

// Some programs
static GLuint showProgram;

static GLuint testProgram;
static GLuint colorProgram;
static GLuint textureProgram;
static GLuint warpProgram;
static GLuint gaussianBlurProgram;
static GLuint bilateralProgram;

static GLuint jointBilateralUpsamplingProgram;

static GLuint histogramProgram;

static GLuint maximaProgram;
static GLuint buildHistoPyramidProgram;
static GLuint compactifyHistoPyramidProgram;
static GLuint featureMarkerProgram;

static GLuint voronoiDiagramProgram;

// A framebuffer
static GLuint framebuffer;

// Some mid-level helper functions

static void getTextureSize(const GLuint texture, const int level, int & width, int &height) {
	glBindTexture(GL_TEXTURE_2D, texture);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, level,  GL_TEXTURE_WIDTH, &width);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, level,  GL_TEXTURE_HEIGHT, &height);
}

// Draw just one test triangle
static void drawTri() {
	glViewport(0, 0, windowWidth, windowHeight);

	glBegin(GL_TRIANGLES) ;  
	glColor4f(0, 1, 1, 0);
	glVertex2f(-0.5f, -0.7f);
	glColor4f(1, 1, 0, 0);
	glVertex2f(0.8f, 0.7f);
	glColor4f(1, 0, 1, 0);
	glVertex2f(-0.2f, 0.8f);  
	glEnd();
}

static void drawFullscreenQuad() {
	glBegin(GL_QUADS) ;
    
	glTexCoord2f(0, 0);
	glVertex2f(-1, -1);

	glTexCoord2f(1, 0);
	glVertex2f(1, -1);

	glTexCoord2f(1, 1);
	glVertex2f(1, 1);

	glTexCoord2f(0, 1);
	glVertex2f(-1, 1);

	glEnd();
}

static void drawPoints(const int width, const int height) {
	glBegin(GL_POINTS) ;
  
	for(int i = 0; i < height; i++) {
		for(int j = 0; j < width; j++) {
			const float x =  float(j) / float(width);
			const float y =  float(i) / float(height);
			glVertex2f(x, y);
		}
	}
  
	glEnd();
}

static void showTexture(const GLuint texture, const int level = 0, const bool useResizing = false) {

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);

	glUseProgramObjectARB(showProgram);  

	if(useResizing) {
		int width, height;
		getTextureSize(texture, level, width, height);
		glViewport(0, 0, width, height);
		setUniformFloat2(showProgram, "windowSize", float(width), float(height));
	} else {
		glViewport(0, 0, windowWidth, windowHeight);
		setUniformFloat2(showProgram, "windowSize", float(windowWidth), float(windowHeight));
	}
	
	setUniformInt(showProgram, "level", level);	

	drawFullscreenQuad();

	glBindTexture(GL_TEXTURE_2D, 0);
}

// Some tests

static void testShader() {

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, lowresTexture);

	//const GLuint program = testProgram;
	//const GLuint program = colorProgram;
	//const GLuint program = textureProgram;
	//const GLuint program = warpProgram;
	//const GLuint program = gaussianBlurProgram;
	//const GLuint program = bilateralProgram;
	//const GLuint program = jointBilateralUpsamplingProgram;
	const GLuint program = maximaProgram;

	glUseProgramObjectARB(program);  

	// Required when using "Color"
	setUniformFloat3(program, "color", 0.5f, 0.1f, 0.2f);
  
	// Required when using an input texture
  	setUniformInt(program, "highresImageSampler2D", 0);
  	setUniformInt(program, "lowresImageSampler2D", 1);

	// Required when using a filter
  	setUniformInt(program, "kernelSize", 40);
	
	int width, height;
	getTextureSize(texture, 0, width, height);

	glViewport(0, 0, width, height);

	drawFullscreenQuad();
}

static void testFBO() {

	// Use this FBO
	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);	

	// Use exactly this test texture
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, fboTestTexture, 0);

	// Use the test program as a test
	glUseProgramObjectARB(testProgram);  
 	
	// Draw, but into a texture
	drawFullscreenQuad();

	// Use the main window framebuffer at handle 0
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	showTexture(fboTestTexture);
}

static void testHistoPyramids() {

	// Use the FBO
	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);	

	// Put result in level 0 of the histoPyramidTexture
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, histoPyramidTexture, 0);

	// Get number of input texture size
	int width, height;
	getTextureSize(texture, 0, width, height);

	// 1st step: Run the maxima finder toproduce level 0
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);

	glUseProgramObjectARB(maximaProgram);  

	glViewport(0, 0, width, height);

	drawFullscreenQuad();	

	// 2nd step: build thepyramid
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, 0, 0);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, histoPyramidTexture);

	glUseProgramObjectARB(buildHistoPyramidProgram);

	const int levelCount = int(floor(logf(float(width)) / logf(2.0f))) + 1;
	cout << levelCount << " levels" << endl;

	for(int i = 1; i < levelCount; i++) {

		const int levelWidth = width >> i;
		const int levelHeight = height>> i;

		cout << "Level " << i << ": " << levelWidth << "x" << levelHeight << endl;
		glViewport(0, 0, levelWidth, levelHeight);

		setUniformInt(buildHistoPyramidProgram, "readLevel", i - 1);

		glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, histoPyramidTexture, i);

		drawFullscreenQuad();
	}

		
	// FILL: 3rd step: Read back the totalCount pixels
	// ...
	// ...
	//const int totalCount = ?;
	//cout << "The total count is " << totalCount << endl;

	// 4th: Compactify the adresses of every sparse maximum

	// This is a mayn pixels as features total
	glViewport(0, 0, 16, 1);
	
	// Write into the result texture
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, resultHistoPyramidTexture, 0);

	// Use the "down" program now
	glUseProgramObjectARB(compactifyHistoPyramidProgram);

	drawFullscreenQuad();

	// Use the main window framebuffer at handle 0
	glBindFramebuffer(GL_FRAMEBUFFER, 0);	

	// Show some slice of the histopyramid to test
	//showTexture(histoPyramidTexture, 4, false);

	// Draw the result, i.e. a list of adresses.
	//showTexture(resultHistoPyramidTexture, 0, false);	

	showTexture(texture);

	// 5th: Draw feature marker
	glUseProgramObjectARB(featureMarkerProgram);

	// FILL: Draw larger points
	// ...
	// ...

	// Draw to the entire output window
	glViewport(0, 0, windowWidth, windowHeight);
	
	// Again, read from the histoPyramid (redundant?)
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, resultHistoPyramidTexture);

	// FILL: Draw as many points as there were featires	
	// ...
	// ...

}

static void testHistogram() {

	// Use this FBO
	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);	

	// Use exactly this test texture
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, histogramTexture, 0);

	// We will use additive blending, so be sure we have 0s in the beginning
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT);

	glUseProgramObjectARB(histogramProgram);  
 	
	// Use the usual texture as the input.
	// Note that the output texture generally has a different size!
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);

	// FILL: Set GL state
	// ...
	// ...

	// FILL: Get the right number of points and draw them to the rig number of output pixels
	// ...
	// ...

	// FILL: reset GL state
	// ...
	// ...

	// Use the main window framebuffer at handle 0
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// Draw it, but scaled, as 1D is to microscopic
	showTexture(histogramTexture, false);
}

static void testVoronoiDiagram() {

	glViewport(0, 0, windowWidth, windowHeight);

	glUseProgramObjectARB(voronoiDiagramProgram);

	// FILL: Set up GL state
	// ...
	// ...
	
	// Do the same every time
	srand(0);

	const int pointCount = 1000;

	// FILL: Draw the points
	// ...
	// ...

	glEnd();

	// FILL: Set back OpenGL state
	// ...
	// ...

	glUseProgramObjectARB(0);
}

static void Display() {
	glClearColor(0.5, 0.5, 0.5, 0.5);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	//drawTri();
	//testShader();	
	//testFBO();
	//testHistogram();
	//testHistoPyramids();
	testVoronoiDiagram();

	glutSwapBuffers();
}

// Global state of the program, global thanks to funny freeglut et al.
bool continue_in_main_loop  = true;

static void Key(unsigned char key, int x, int y) {
  switch (key) {
  case 27:
    continue_in_main_loop = false;
    break;
  }
}

static void Reshape(int width, int height) {  
	windowWidth = width;
	windowHeight = height;
}

int main(int argc, char *argv[]) {

	PVC_Image* const image = loadImage("../Data/stars.ppm");	

	glutInitWindowSize(image->w(), image->h());
	glutInitWindowPosition(200, 200);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH);
	glutInit(&argc, argv);

	glutCreateWindow( "OpenGL Test" );

	const GLenum glewError = glewInit();
	if (GLEW_OK != glewError) {
		cerr << "Error:" << glewGetErrorString(glewError);
	}
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	// Create one framebuffer, thats enough for our case
	glGenFramebuffers(1, &framebuffer);

	texture = loadTexture(image);
	
	// Helper program
	showProgram = loadProgram("Show");

	// --- 1st lecture -------------------------------
	//testProgram = loadProgram("Test");
	//colorProgram = loadProgram("Color");
	//textureProgram = loadProgram("Texture");
	//warpProgram = loadProgram("Warp");
	//gaussianBlurProgram = loadProgram("GaussianBlur");
	//bilateralProgram = loadProgram("Bilateral");
	
	//jointBilateralUpsamplingProgram = loadProgram("JointBilateralUpsampling");
	//lowresTexture = loadTexture(loadImage("../Data/depth lowres.ppm"));

	// --- 2nd lecture -------------------------------
	fboTestTexture = allocateTexture(256, 256);	

	// Assignment 5.1.
	histogramProgram = loadProgram("Histogram");
	
	histogramTexture = allocateTexture(256, 1, GL_LUMINANCE32F_ARB);
	
	// Assignment 5.2.
	maximaProgram = loadProgram("Maxima");
	buildHistoPyramidProgram = loadProgram("BuildHistoPyramid");
	compactifyHistoPyramidProgram = loadProgram("CompactifyHistoPyramid");
	featureMarkerProgram = loadProgram("FeatureMarker");
	
	histoPyramidTexture = allocateTexture(image->w(), image->h(), GL_LUMINANCE32F_ARB, true);	
	resultHistoPyramidTexture = allocateTexture(16, 1, GL_RGB32F_ARB);	

	// Assignment 5.3.
	voronoiDiagramProgram = loadProgram("VoronoiDiagram");

	glutKeyboardFunc(Key);
	glutReshapeFunc(Reshape);
	glutDisplayFunc(Display);

	while(continue_in_main_loop) {
		glutMainLoopEvent();
	}
}
